from . import segmentation, sc_reference, deconvolution, plot
from .sc_reference import (
    initialization,
    marker_selection,
    construct_sc_ref,
    plot_sc_ref,
)
from .deconvolution import deconvolute, simulation

__all__ = [
    "segmentation",
    "initialization",
    "marker_selection",
    "construct_sc_ref",
    "plot_sc_ref",
    "deconvolute",
    "simulation",
]



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from stardist.models import StarDist2D
from csbdeep.utils import normalize
import tensorflow as tf
from packaging import version
import os
from scipy.signal import convolve2d
from tqdm import tqdm
import cv2
from typing import List, Dict, Tuple, Union

os.environ["CUDA_LAUNCH_BLOCKING"] = "1"


class Segmentation:
    def __init__(
        self,
        img: np.ndarray,
        spot_center: np.ndarray,
        out_dir: str = "",
        prob_thresh: float = 0.2,
        nms_thresh: float = 0.5,
        spot_radius: float = 36.5,
        n_tiles: Tuple[int, int, int] = (8, 8, 1),
        enhancement: bool = False,
        enhancement_params: dict = None,
    ):
        """
        Args:
            img (numpy.ndarray):. Three channel stained image. In default, it should be hematoxylin and eosin (H&E) stained
                image.
            spot_center: Coordinates of the center of the spots.
            out_dir: Output directory.
            Prob_thresh, nms_thresh: Two thresholds used in Stardist. User should adjust these two threshold based on
                the segmentation results.
            spot_radius: Radius of the spots. In 10X Visium, it should be 36.5.
            n_tiles: Out of memory (OOM) errors can occur if the input image is too large. To avoid this problem, the
                input image is broken up into (overlapping) tiles that are processed independently and re-assembled.
                (Copied from stardist document). In default, we break the image into 8*8 tiles.
            enhancement: Whether to apply image enhancement before segmentation. (testing)
            enhancement_params: Optional, the parameter in the enhancement procedure. If not provided, use default. (testing)

        """
        self.enhancement_params = (
            enhancement_params if enhancement_params else {"cla_wt": 0.2}
        )
        if enhancement:
            self.img = self.apply_enhancement(
                img, cla_wt=self.enhancement_params.get("cla_wt", 0.2)
            )
        else:
            self.img = img

        if spot_center is not None:
            if spot_center.shape[1] != 2:
                raise ValueError("spot_center must have exactly two columns.")
        self.spot_center = np.array(spot_center)
        self.prob_thresh = prob_thresh
        self.nms_thresh = nms_thresh
        self.spot_radius = spot_radius
        if out_dir and not os.path.exists(out_dir):
            os.mkdir(out_dir)
        self.out_dir = out_dir + "/" if out_dir else ""
        # Segmentation results
        self.label, self.nucleus_boundary, self.probability = None, None, None
        self.n_cell_df = None
        self.nucleus_df = None
        self.is_segmented = False  # Whether we have conducted segmentation or not.
        self.n_tiles = n_tiles
        if version.parse(tf.__version__) >= version.parse("2.9.0"):
            tf.keras.Model.predict = change_predict_defaults(tf.keras.Model.predict)
            print(
                f"Suppress the output of tensorflow prediction for tensorflow version {tf.version.VERSION}>=2.9.0."
            )

    @staticmethod
    def apply_enhancement(img: np.ndarray, cla_wt: float) -> np.ndarray:
        """Apply CLAHE and bilateral filtering to enhance the image.

        Args:
            img: Input RGB image.
            cla_wt: Weight for combining CLAHE and bilateral filtering results. Must be between 0 and 1.

        Returns:
            np.ndarray: Enhanced image.
        """

        def _bilateral(img):
            bilateral_filtered = cv2.bilateralFilter(
                img, d=9, sigmaColor=75, sigmaSpace=75
            )
            return bilateral_filtered

        def _clahe(img):
            lab = cv2.cvtColor(img, cv2.COLOR_RGB2LAB)
            l, a, b = cv2.split(lab)
            clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
            l_clahe = clahe.apply(l)
            lab_clahe = cv2.merge((l_clahe, a, b))
            result_rgb = cv2.cvtColor(lab_clahe, cv2.COLOR_LAB2RGB)
            return result_rgb

        def _CLA_BIL(img, cla_wt):
            clahe_img = _clahe(img)
            bil_img = _bilateral(img)
            combined_result = cv2.addWeighted(clahe_img, cla_wt, bil_img, 1 - cla_wt, 0)
            return combined_result

        img = _CLA_BIL(img, cla_wt)

        return img

    @staticmethod
    def stardist_2D_versatile_he(
        img: np.ndarray,
        prob_thresh: float = 0.2,
        nms_thresh: float = 0.5,
        n_tiles: Tuple[int, int, int] = (8, 8, 1),
        verbose: bool = True,
    ) -> Tuple[np.ndarray, dict]:
        """Segmentation function provided by Stardist.

        Args:
            img: Three channel image.
            nms_thresh: Parameter of non-maximum suppression.
            prob_thresh: The probability threshold that determines the retention of a nucleus.
            n_tiles: Out of memory (OOM) errors can occur if the input image is too large. To avoid this problem, the
                input image is broken up into (overlapping) tiles that are processed independently and re-assembled.
                (Copied from stardist document). In default, we break the image into 8*8 tiles.
            verbose: Whether to print segmentation progress.

        Returns:
            There are two returns.

            np.ndarray: The segmented image. Background pixels has value 0 and nucleus pixels has
            the positive integer value as the index of the nucleus.

            list: Nucleus details. Details[0]: np.ndarray(n_nucleus*2*32). Boundaries of each nucleus. Details[1]:
            np.ndarray(n_nucleus*2). Center of each nucleus. Details[2]: np.ndarray(n_nucleus). Probability that a
            segmented nucleus is indeed a nucleus.
        """

        axis_norm = (0, 1, 2)  # normalize channels jointly
        img = normalize(img, 1, 99.8, axis=axis_norm)
        model = StarDist2D.from_pretrained("2D_versatile_he")
        label, details = model.predict_instances(
            img,
            nms_thresh=nms_thresh,
            prob_thresh=prob_thresh,
            n_tiles=n_tiles,
            show_tile_progress=verbose,
        )
        return label, details

    @staticmethod
    def n_cell_in_spot(
        nucleus_center: np.ndarray,
        spot_center: np.ndarray,
        spot_radius: float,
        nucleus_df: pd.DataFrame = None,
    ) -> pd.DataFrame:
        """Find the number of cells in each spot.

        If the center of a nucleus is inside the spot, we assume that the cell is in the spot.

        Args:
            nucleus_center: np.ndarray(n_nucleus*2). Coordinates of the nucleus centers.
            spot_center: np.ndarray(n_spot*2). Coordinates of the spot centers.
            spot_radius: Radius of the spots.
            nucleus_df: Optional, dataframe of the nucleus.

        Returns:
            Pandas data frame with two columns. Column 'cell_count' represents the number of cells in a spot.
            Column 'centers' represents the coordinates of the nucleus centers.
        """
        n_spot = len(spot_center)
        n_cell_df = pd.DataFrame(
            {
                "cell_count": [0] * n_spot,
                "Nucleus centers": None,
                "Nucleus indices": None,
            }
        )
        distance = np.sum(
            (spot_center[:, :, np.newaxis] - nucleus_center.T) ** 2, axis=1
        )
        if nucleus_df is not None:
            nucleus_df["in_spot"] = False
        for i in range(n_spot):
            nucleus_index = np.where(distance[i] < spot_radius**2)[0]
            n_cell_df.iloc[i, 0] = len(nucleus_index)
            n_cell_df.at[i, "Nucleus centers"] = (
                nucleus_center[nucleus_index] if len(nucleus_index) else np.array([])
            )
            n_cell_df.at[i, "Nucleus indices"] = nucleus_index
            if nucleus_df is not None:
                nucleus_df.loc[nucleus_index, "in_spot"] = True
        return n_cell_df

    def segment_nucleus(self, save: bool = True) -> None:
        """Conduct the segmentation using StarDist pretrained model.

        Args:
            save: Whether to save the segmentation results.
        """
        self.label, details = self.stardist_2D_versatile_he(
            self.img,
            nms_thresh=self.nms_thresh,
            prob_thresh=self.prob_thresh,
            n_tiles=self.n_tiles,
        )
        nucleus_boundary, nucleus_center, self.probability = (
            details["coord"],
            details["points"],
            details["prob"],
        )
        nucleus_boundary = np.transpose(nucleus_boundary, [0, 2, 1])
        self.nucleus_boundary = nucleus_boundary[:, :, [1, 0]]
        self.nucleus_df = pd.DataFrame(
            {"x": nucleus_center[:, 1], "y": nucleus_center[:, 0]}
        )
        self.n_cell_df = self.n_cell_in_spot(
            self.nucleus_df[["x", "y"]].values,
            self.spot_center,
            self.spot_radius,
            self.nucleus_df,
        )
        self.is_segmented = True
        if save:
            self.save_results()

    def save_results(self) -> None:
        """Save segmentation results."""
        assert self.is_segmented, "Please conduct segmentation first."
        np.save(f"{self.out_dir}segmentation_label.npy", self.label)
        np.save(f"{self.out_dir}segmentation_boundary.npy", self.nucleus_boundary)
        np.save(f"{self.out_dir}nucleus_df.csv", self.nucleus_df)
        np.save(f"{self.out_dir}segmentation_probability.npy", self.probability)
        self.n_cell_df.to_csv(f"{self.out_dir}n_cell_df.csv")

    def plot(
        self,
        fig_size: Tuple[int, float] = (10, 4.5),
        dpi: int = 300,
        crop: Union[Tuple[int, int, int, int], None] = None,
        cmap_segmented: str = "hot",
        save: bool = False,
        path: str = None,
    ) -> None:
        """Plot the segmentation results.

        It is recommended to adjust the stardist parameters nms_thresh and prob_thresh based on this plot.

        Args:
            fig_size: Size of the figure.
            dpi: Dots per inch (DPI) of the figure.
            crop: If None, show the full image. Otherwise, crop should be
            cmap_segmented: Color map of the segmented image.
            save: If true, save the figure.
            path: Path to the save figure.
        """
        assert self.is_segmented, "Please conduct segmentation first."
        fig, ax = plt.subplots(1, 2, figsize=fig_size, dpi=dpi)
        if crop is None:
            img = self.img
            img_segmented = self.label
        else:
            img = self.img[crop[0] : crop[1], crop[2] : crop[3]]
            img_segmented = self.label[crop[0] : crop[1], crop[2] : crop[3]]
        ax[0].imshow(img)
        ax[0].set_title("Original image")
        ax[1].imshow(img_segmented, cmap=cmap_segmented)
        ax[1].set_title("Segmented image")
        ax[0].axis("off")
        ax[1].axis("off")
        if save:
            plt.savefig(path, bbox_inches="tight")
        plt.show()


def change_predict_defaults(predict_function):
    """Wrap prediction function for tensorflow>=2.9.0 to reduce outputs.

    Start from tensorflow 2.9.0, the default verbose value in function tensorflow.keras.Model.predict is set to 'auto',
    which takes value 1 for the most of the time. Thus, this function is a decorator to wrap the predict function
    and set the default verbose to 0.

    Args:
        predict_function: tensorflow.keras.Model.predict
    """

    def wrapper(instance, x, verbose=0, **kwargs):
        return predict_function(instance, x, verbose=verbose, **kwargs)

    return wrapper


def cell_boundary(
    nucleus_location: np.ndarray,
    img_size: Tuple[int, int],
    max_dist: float,
    max_area: float,
    search_direction: List[List[int]],
    verbose: int = 0,
    delta: int = 1,
) -> dict:
    """
    Infer the cell boundary.

    Args:
        nucleus_location: Coordinates of the nucleus centers.
        img_size: The size of the image.
        max_dist: The largest distance from the cell boundary to the nucleus center.
        max_area: Largest area of a cell. Segmented nuclei with size larger than this value are discarded.
        search_direction: The search direction when determine the cell boundary.
        verbose: Whether to print progress.
        delta: Increment of the radius in each round.

    Returns:
        Dictionary of the cell boundary information.
    """
    img_cell = np.zeros(img_size)
    n_nuclei = len(nucleus_location)
    n_pixel = [0] * n_nuclei
    nuclei_left = set(range(1, n_nuclei + 1))

    radius = 0
    while nuclei_left and radius < max_dist:
        radius += delta
        if verbose:
            print(
                "r: "
                + str(radius)
                + "\tThere are "
                + str(len(nuclei_left))
                + " nuclei left."
            )
        pixel_check = []
        for dx in range(-radius - 1, radius + 2):
            for dy in range(-radius - 1, radius + 2):
                if (radius - delta) ** 2 <= dx**2 + dy**2 <= radius**2:
                    pixel_check += [[dx, dy]]

        nuclei_remove = set()
        for i in nuclei_left:
            border_temp = []
            c_y = int(nucleus_location[i - 1, 0])
            c_x = int(nucleus_location[i - 1, 1])
            flag = True  # Whether the expended space are all occupied by background or other cells.
            for dx, dy in pixel_check:
                x1 = c_x + dx
                y1 = c_y + dy
                if (
                    0 <= x1 <= img_cell.shape[0] - 1
                    and 0 <= y1 <= img_cell.shape[1] - 1
                    and dx**2 + dy**2 <= radius**2
                ):
                    if img_cell[x1, y1] == 0:
                        img_cell[x1, y1] = i
                        border_temp += [[x1, y1]]
                    if img_cell[x1, y1] == i:
                        flag = False
            n_pixel[i - 1] += len(border_temp)
            if n_pixel[i - 1] > max_area or (len(border_temp) == 0 and flag):
                nuclei_remove.add(i)
        for i in nuclei_remove:
            nuclei_left.remove(i)

    img_cell_boundaries = np.zeros((img_cell.shape[0], img_cell.shape[1], 3))
    individual_boundary = {i: [] for i in range(n_nuclei + 1)}

    search_direction = np.array(search_direction)
    boundary_img = np.zeros_like(img_cell)
    for dx, dy in tqdm(search_direction):
        d = max([abs(dx), abs(dy)])
        kernel = np.zeros((2 * d + 1, 2 * d + 1))
        kernel[dy + d, dx + d] = 1
        kernel[d, d] = -1
        convolved = convolve2d(img_cell, kernel, mode="same")
        boundary_img += np.where(convolved != 0, 1, 0)
    img_cell_boundaries[boundary_img > 0] = [150, 150, 150]
    for x in tqdm(range(len(img_cell))):
        for y in range(len(img_cell[0])):
            if boundary_img[x, y] > 0:
                individual_boundary[img_cell[x, y]] += [[x, y]]

    img_cell_boundaries = img_cell_boundaries.astype(np.int32)
    individual_boundary = {
        i: np.array(individual_boundary[i]) for i in range(n_nuclei + 1)
    }
    d = {
        "img_cell": img_cell,
        "cell_boundary": img_cell_boundaries,
        "size": n_pixel,
        "individual_boundary": individual_boundary,
    }
    return d


def add_boundary(img: np.ndarray, boundary: np.ndarray) -> np.ndarray:
    """
    Add boundary overlay to image.
    Args:
        img: Input image (np.ndarray).
        boundary: Boundary mask (np.ndarray).
    Returns:
        Image with boundary overlay.
    """
    img = img.copy()
    if img.shape[:2] != boundary.shape[:2]:
        raise ValueError("Image and boundary must have the same spatial dimensions.")
    img[boundary[..., 0] > 0] = [50, 50, 50]
    return img



import anndata
import os
import cv2 as cv
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import pandas as pd
import seaborn as sns
import numpy as np
import torch
import pyro
from tqdm import tqdm
import pyro.distributions as dist
import pyro.distributions.constraints as constraints
from pyro.optim import Adam
from pyro.infer import SVI, Trace_ELBO
from scipy.spatial.distance import jensenshannon
import time
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel as C
from spotiphy import sc_reference
from typing import List, Dict, Tuple, Union, Optional


def deconvolute(
    X: np.ndarray,
    sc_ref: np.ndarray,
    device: str = "cuda",
    n_epoch: int = 8000,
    adam_params: Optional[Dict] = None,
    batch_prior: float = 2,
    plot: bool = False,
    fig_size: Tuple[float, float] = (4.8, 3.6),
    dpi: int = 200,
):
    """Deconvolution of the proportion of genes contributed by each cell type.

    Args:
        X: Spatial transcriptomics data. n_spot*n_gene.
        sc_ref: Single cell reference. n_type*n_gene.
        device: The device used for the deconvolution.
        plot: Whether to plot the ELBO loss.
        n_epoch: Number of training epochs.
        adam_params: Parameters for the adam optimizer.
        batch_prior: Parameter of the prior distribution of the batch effect: 2^(Uniform(0, batch_prior))
        fig_size: Size of the figure.
        dpi: Dots per inch (DPI) of the figure.

    Returns:
        Parameters in the generative model.
    """
    if adam_params is None:
        adam_params = {"lr": 0.003, "betas": (0.95, 0.999)}
    max_exp = int(np.max(np.sum(X, axis=1)))
    X = torch.tensor(X, device=device, dtype=torch.int32)
    sc_ref = torch.tensor(sc_ref, device=device)
    batch_prior = torch.tensor(batch_prior, device=device, dtype=torch.float64)
    assert (
        X.shape[1] == sc_ref.shape[1]
    ), "Spatial data and SC reference data must have the same number of genes."

    def model(sc_ref, X, batch_prior):
        n_spot = len(X)
        n_type, n_gene = sc_ref.shape
        # alpha = pyro.sample("Batch effect", dist.Dirichlet(torch.full((n_gene,), batch_prior, device=device,
        #                                                               dtype=torch.float64)))
        alpha = pyro.sample(
            "Batch effect",
            dist.Uniform(
                torch.tensor(0.0, device=device), torch.tensor(1.0, device=device)
            )
            .expand([n_gene])
            .to_event(1),
        )
        alpha_exp = torch.exp2(alpha * batch_prior)
        with pyro.plate("spot", n_spot, dim=-1):
            q = pyro.sample(
                "Proportion",
                dist.Dirichlet(
                    torch.full((n_type,), 3.0, device=device, dtype=torch.float64)
                ),
            )
            rho = torch.matmul(q, sc_ref)
            rho = rho * alpha_exp
            # print(torch.min(rho.sum(dim=-1).unsqueeze(-1)))
            rho = rho / rho.sum(dim=-1).unsqueeze(-1)
            if torch.any(torch.isnan(rho)):
                print("rho contains NaN at")
                print(torch.where(torch.isnan(rho)))
            pyro.sample(
                "Spatial RNA", dist.Multinomial(total_count=max_exp, probs=rho), obs=X
            )

    def guide(sc_ref, X, batch_prior):
        n_spot = len(X)
        n_type, n_gene = sc_ref.shape
        # pi = pyro.param('pi', lambda: torch.full((n_gene,), 10, device=device, dtype=torch.float64),
        #                 constraint=constraints.positive)
        # alpha = pyro.sample("Batch effect", dist.Dirichlet(pi))
        alpha_loc = pyro.param(
            "alpha_loc", torch.zeros(n_gene, device=device, dtype=torch.float64)
        )
        alpha_scale = pyro.param(
            "alpha_scale",
            torch.ones(n_gene, device=device, dtype=torch.float64),
            constraint=constraints.positive,
        )
        base_dist = dist.Normal(alpha_loc, alpha_scale)
        sigmoid_trans = dist.transforms.SigmoidTransform()  # Transforms to [0, 1]

        alpha_dist = pyro.distributions.TransformedDistribution(
            base_dist, [sigmoid_trans]
        )

        alpha = pyro.sample("Batch effect", alpha_dist.to_event(1))
        with pyro.plate("spot", n_spot):
            sigma = pyro.param(
                "sigma",
                lambda: torch.full(
                    (n_spot, n_type), 3.0, device=device, dtype=torch.float64
                ),
                constraint=constraints.positive,
            )
            if torch.any(torch.isnan(sigma)):
                print("sigma contains NaN at")
                print(torch.where(torch.isnan(sigma)))
            # print(torch.min(sigma))
            q = pyro.sample("Proportion", dist.Dirichlet(sigma))

    pyro.clear_param_store()
    optimizer = Adam(adam_params)
    svi = SVI(model, guide, optimizer, loss=Trace_ELBO())
    loss_history = []
    for _ in tqdm(range(n_epoch)):
        loss = svi.step(sc_ref, X, batch_prior)
        loss_history.append(loss)
    if plot:
        plt.figure(figsize=fig_size, dpi=dpi)
        plt.plot(loss_history, lw=0.5)
        plt.ylabel("Loss")
        plt.xlabel("Epoch")
        plt.show()
    return pyro.get_param_store()


def estimation_proportion(
    X: np.ndarray,
    adata_sc: anndata.AnnData,
    sc_ref: np.ndarray,
    type_list: List[str],
    key_type: str,
    device: str = "cuda",
    n_epoch: int = 8000,
    adam_params: Optional[Dict] = None,
    batch_prior: float = 2,
    plot: bool = False,
    fig_size: Tuple[float, float] = (4.8, 3.6),
    dpi: int = 200,
) -> np.ndarray:
    """
    Estimate the proportion of each cell type in each spot.

    Args:
        X: Spatial transcriptomics data. n_spot*n_gene.
        adata_sc (anndata.Anndata): scRNA data.
        sc_ref: Single cell reference. n_type*n_gene.
        type_list: List of the cell types.
        key_type: Column name of the cell types in adata_sc.
        device: The device used for the deconvolution.
        plot: Whether to plot the ELBO loss.
        n_epoch: Number of training epochs.
        adam_params: Parameters for the adam optimizer.
        batch_prior: Parameter of the prior Dirichlet distribution of the batch effect: 2^(Uniform(0, batch_prior))
        fig_size: Size of the figure.
        dpi: Dots per inch (DPI) of the figure.

    Returns:
        Parameters in the generative model.
    """
    pyro_params = deconvolute(
        X,
        sc_ref,
        device=device,
        n_epoch=n_epoch,
        adam_params=adam_params,
        batch_prior=batch_prior,
        plot=plot,
        fig_size=fig_size,
        dpi=dpi,
    )
    sigma = pyro_params["sigma"].cpu().detach().numpy()
    Y = np.array(adata_sc.X)
    mean_exp = np.array(
        [
            np.mean(np.sum(Y[adata_sc.obs[key_type] == type_list[i]], axis=1))
            for i in range(len(type_list))
        ]
    )
    cell_proportion = sigma / mean_exp
    cell_proportion = cell_proportion / np.sum(cell_proportion, axis=1)[:, np.newaxis]
    return cell_proportion


def plot_proportion(
    img: np.ndarray,
    proportion: np.ndarray,
    spot_location: np.ndarray,
    radius: float,
    cmap_name: str = "viridis",
    alpha: float = 0.4,
    save_path: Optional[str] = "proportion.png",
    vmax: float = 0.98,
    spot_scale: float = 1.3,
    show_figure: bool = False,
    int_ticks: bool = False,
    bar_location: Tuple[int, int] = (5800, 8100),
) -> None:
    """
    Plot the proportion of a cell type.

    Args:
        img: 3 channel img with integer values in [0, 255]
        proportion: Proportion of a cell type.
        spot_location: Location of the spots.
        radius: Radius of the spot
        cmap_name: Name of the camp.
        alpha: Level of transparency of the background img.
        save_path: If not none, save the img to the path.
        vmax: Quantile of the maximum value in the color bar.
        spot_scale: Scale of the spot in the figure.
        show_figure: Whether plot the figure.
        int_ticks: Whether the ticks must be integers.
    """

    def render_to_array(fig):
        fig.canvas.draw()
        buf = fig.canvas.tostring_rgb()
        ncols, nrows = fig.canvas.get_width_height()
        return np.frombuffer(buf, dtype=np.uint8).reshape(nrows, ncols, 3)

    img = img * alpha + np.ones(img.shape) * 255 * (1 - alpha)
    img = img.astype(np.int32)
    vmin = 0
    vmax = np.quantile(proportion, vmax)
    if vmax - vmin < 0.03:
        vmax = vmin + 0.03
    proportion = (proportion - vmin) / vmax
    cmap = plt.cm.get_cmap(cmap_name)
    for i, p in enumerate(proportion):
        rgb_float = cmap(p)[:3]
        rgb_int = tuple(int(255 * x) for x in rgb_float)
        cv.circle(img, spot_location[i], int(radius * spot_scale), rgb_int, -1)

    fig = plt.figure(figsize=(1, 2.5), dpi=1200)
    cbar_ax = fig.add_axes([0, 0.1, 0.25, 0.85])
    if vmax >= 0.2:
        a = int(vmax * 10) / 20
    else:
        a = int(vmax * 100) / 200
    if int_ticks:
        a = int(a)
    ticks = [0, a, a * 2]
    cb1 = plt.colorbar(
        plt.cm.ScalarMappable(norm=plt.Normalize(vmin, vmax), cmap=cmap),
        cax=cbar_ax,
        ticks=ticks,
    )
    cb1.outline.set_edgecolor("none")
    for label in cb1.ax.get_yticklabels():
        label.set_size(13)
    cbar_array = render_to_array(fig)
    plt.close(fig)

    c = [bar_location[0], bar_location[1]]
    img[c[0] : c[0] + cbar_array.shape[0], c[1] : c[1] + cbar_array.shape[1]][
        np.sum(cbar_array, axis=2) < 252 * 3
    ] = cbar_array[np.sum(cbar_array, axis=2) < 252 * 3]

    if save_path is not None:
        print("Saving the image.")
        img1 = img[:, :, [2, 1, 0]]
        cv.imwrite(save_path, img1)
    plt.imshow(img)
    if not show_figure:
        plt.close()


def proportion_to_count(
    p: np.ndarray, n: Union[int, np.ndarray], multiple_spots: bool = False
) -> np.ndarray:
    """
    Convert the cell proportion to the absolute cell number.

    Args:
        p: Cell proportion.
        n: Number of cells.
        multiple_spots: If the data is related to multiple spots

    Returns:
        Cell count of each cell type.
    """
    if multiple_spots:
        assert len(p) == len(n)
        count = np.zeros(p.shape)
        for i in range(len(p)):
            count[i] = proportion_to_count(p[i], n[i])
    else:
        assert np.abs(np.sum(p) - 1) < 1e-5
        c0 = p * n
        count = np.floor(c0)
        r = c0 - count
        if np.sum(count) == n:
            return count
        idx = np.argsort(r)[-int(np.round(n - np.sum(count))) :]
        count[idx] += 1
    return count


def simulation(
    adata_st: anndata.AnnData,
    adata_sc: anndata.AnnData,
    key_type: str,
    cell_proportion: np.ndarray,
    n_cell: Union[int, str] = 10,
    batch_effect_sigma: float = 0.1,
    zero_proportion: float = 0.3,
    additive_noise_sigma: float = 0.05,
    save: bool = True,
    out_dir: str = "",
    filename: str = "ST_Simulated.h5ad",
    verbose: int = 0,
) -> anndata.AnnData:
    """
    Simulation of the spatial transcriptomics data based on a real spatial sample and deconvolution results of Spotiphy.

    Args:
        adata_st: Original spatial transcriptomics data.
        adata_sc: Original single-cell data.
        key_type: The key that is used to extract cell type information from adata_sc.obs.
        cell_proportion: Proportion of each cell type obtained by the deconvolution.
        n_cell: Number of cells in each spot, either a key of adata_st.obs or a positive integer.
        batch_effect_sigma: Sigma of the log-normal distribution when generate batch effect.
        zero_proportion: Proportion of gene expression set to 0. Note that since some gene expression in the original
            X is already 0, the final proportion of 0 gene read is larger than zero_proportion.
        additive_noise_sigma: Sigma of the log-normal distribution when generate additive noise.
        save: If True, save the generated adata_st as a file.
        out_dir: Output directory.
        filename: Name of the saved file.
        verbose: Whether print the time spend.
    Returns:
        Simulated ST Anndata.
    """
    time_start = time.time()
    # Construct ground truth
    type_list = sorted(list(adata_sc.obs[key_type].unique()))  # list of the cell type.
    assert len(type_list) == cell_proportion.shape[1]
    assert len(cell_proportion) == len(adata_st)
    n_spot = len(adata_st)
    n_type = len(type_list)
    if isinstance(n_cell, (int, float)):
        assert n_cell >= 1
        n_cell = np.array([int(n_cell)] * n_spot)
    else:
        n_cell = adata_st.obs[n_cell].values.astype(int)
        n_cell[n_cell <= 0] = 1
    cell_count = np.zeros(cell_proportion.shape)
    for i in range(n_spot):
        cell_count[i] = proportion_to_count(cell_proportion[i], n_cell[i])
    cell_count = cell_count.astype(int)
    adata_st.obsm["ground_truth"] = cell_count
    adata_st.obs["cell_count"] = n_cell
    adata_st.uns["type_list"] = type_list
    if verbose:
        print(
            "Prepared the ground truth. Time use {:.2f}".format(
                time.time() - time_start
            )
        )
        time_start = time.time()

    # Construct expression matrix
    common_genes = list(set(adata_st.var_names).intersection(set(adata_sc.var_names)))
    adata_sc = adata_sc[:, common_genes]
    adata_st = adata_st[:, common_genes]
    n_gene = len(common_genes)
    X = np.zeros(adata_st.shape)
    Y = adata_sc.X if type(adata_sc.X) is np.ndarray else adata_sc.X.toarray()
    Y = Y * 1e6 / np.sum(Y, axis=1, keepdims=True)
    type_index = []
    for i in range(n_type):
        type_index.append(np.where(adata_sc.obs[key_type] == type_list[i])[0])
    for i in range(n_spot):
        for j in range(n_type):
            if cell_count[i, j] > 0:
                X_temp = np.array(
                    np.sum(Y[np.random.choice(type_index[j], cell_count[i, j])], axis=0)
                )
                if X_temp.ndim > 1:
                    X[i] += X_temp[0]
                else:
                    X[i] += X_temp
    if verbose:
        print(
            "Constructed the ground truth. Time use {:.2f}".format(
                time.time() - time_start
            )
        )
        time_start = time.time()

    # Add noise
    # Batch effect
    batch_effect = np.random.lognormal(0, batch_effect_sigma, size=n_gene)
    X = X * batch_effect
    # Zero reads
    zero_read = np.random.binomial(1, 1 - zero_proportion, size=X.shape)
    X = X * zero_read
    # Additive noise
    additive_noise = np.random.lognormal(0, additive_noise_sigma, size=X.shape)
    X = X * additive_noise
    adata_st.X = X
    adata_st.uns["batch_effect"] = batch_effect
    if verbose:
        print(
            "Added batch effect, zero reads, and additive noise. Time use {:.2f}".format(
                time.time() - time_start
            )
        )
        time_start = time.time()

    if save:
        if out_dir and not os.path.exists(out_dir):
            os.mkdir(out_dir)
        out_dir = out_dir + "/" if out_dir else ""
        adata_st.raw = None
        adata_st.write(out_dir + filename)
        if verbose:
            print(
                "Saved the simulated data to file. Time use {:.2f}".format(
                    time.time() - time_start
                )
            )

    return adata_st


class Evaluation:
    def __init__(
        self,
        proportion_truth: np.ndarray,
        proportion_estimated_list: List[np.ndarray],
        methods: List[str],
        out_dir: str = "",
        cluster: Optional[np.ndarray] = None,
        type_list: Optional[np.ndarray] = None,
        colors: Optional[List[str]] = None,
        coordinates: Optional[np.ndarray] = None,
        min_spot_distance: int = 112,
    ):
        """
        Args:
            proportion_truth: Ground truth of the cell proportion.
            proportion_estimated_list: List of estimated proportion by each method.
            methods: List of methods names.
            out_dir: Output directory.
            cluster: Cluster label of each spot.
            type_list: List of cell types
        """
        self.proportion_truth = proportion_truth
        self.proportion_estimated_list = proportion_estimated_list
        self.methods = methods
        self.metric_dict = dict()  # Saved metric values.
        self.n_method = len(methods)
        self.coordinates = coordinates
        self.min_spot_distance = min_spot_distance
        self.cluster = cluster
        self.type_list = type_list
        self.n_type = proportion_truth.shape[1]

        self.function_map = {
            "Absolute error": self.absolute_error,
            "Square error": self.square_error,
            "Cosine similarity": self.cosine,
            "Correlation": self.correlation,
            "JSD": self.JSD,
            "Fraction of cells correctly mapped": self.correct_fraction,
        }
        self.metric_type_dict = {
            "Spot": {
                "Cosine similarity",
                "Absolute error",
                "Square error",
                "JSD",
                "Correlation",
                "Fraction of cells correctly mapped",
            },
            "Cell type": {
                "Cosine similarity",
                "Absolute error",
                "Square error",
                "JSD",
                "Fraction of cells correctly mapped",
                "Correlation",
            },
            "Individual": {"Absolute error", "Square error"},
        }
        if colors is None:
            self.colors = [
                "#3c93c2",
                "#089099",
                "#7ccba2",
                "#fcde9c",
                "#f0746e",
                "#dc3977",
                "#7c1d6f",
            ]
        else:
            self.colors = colors
        self.colors = self.colors[: len(self.methods)]

        if out_dir:
            self.out_dir = out_dir if out_dir[-1] == "/" else out_dir + "/"
        else:
            self.out_dir = ""
        if self.out_dir and not os.path.exists(self.out_dir):
            os.mkdir(self.out_dir)
        if not os.path.exists(out_dir + "figures"):
            os.mkdir(out_dir + "figures")
        assert len(proportion_estimated_list) == self.n_method

    @staticmethod
    def absolute_error(
        proportion_truth: np.ndarray,
        proportion_estimated: np.ndarray,
        metric_type: str = "Spot",
    ) -> np.ndarray:
        assert proportion_truth.shape == proportion_estimated.shape
        error = np.abs(proportion_truth - proportion_estimated)
        if metric_type == "Individual":
            return error
        elif metric_type == "Spot":
            return np.sum(error, axis=1)
        elif metric_type == "Cell type":
            return np.sum(error, axis=0)
        else:
            raise ValueError(f"Invalid metric type {metric_type}")

    @staticmethod
    def square_error(
        proportion_truth: np.ndarray,
        proportion_estimated: np.ndarray,
        metric_type: str = "Spot",
    ) -> np.ndarray:
        assert proportion_truth.shape == proportion_estimated.shape
        error = (proportion_truth - proportion_estimated) ** 2
        if metric_type == "Individual":
            return error
        elif metric_type == "Spot":
            return np.sum(error, axis=1)
        elif metric_type == "Cell type":
            return np.sum(error, axis=0)
        else:
            raise ValueError(f"Invalid metric type {metric_type}")

    @staticmethod
    def cosine(
        proportion_truth: np.ndarray,
        proportion_estimated: np.ndarray,
        metric_type: str = "Spot",
    ) -> np.ndarray:
        assert proportion_truth.shape == proportion_estimated.shape
        if metric_type == "Spot":
            cosine_similarity = (
                np.sum(proportion_truth * proportion_estimated, axis=1)
                / np.linalg.norm(proportion_estimated, axis=1)
                / np.linalg.norm(proportion_truth, axis=1)
            )
            return cosine_similarity
        elif metric_type == "Cell type":
            cosine_similarity = (
                np.sum(proportion_truth * proportion_estimated, axis=0)
                / np.linalg.norm(proportion_estimated, axis=0)
                / np.linalg.norm(proportion_truth, axis=0)
            )
            return cosine_similarity
        else:
            raise ValueError(f"Invalid metric type {metric_type}")

    @staticmethod
    def correlation(
        proportion_truth: np.ndarray,
        proportion_estimated: np.ndarray,
        metric_type: str = "Spot",
    ) -> np.ndarray:
        assert proportion_truth.shape == proportion_estimated.shape
        if metric_type == "Cell type":
            proportion_truth_centered = proportion_truth - np.mean(
                proportion_truth, axis=0
            )
            proportion_estimated_centered = proportion_estimated - np.mean(
                proportion_estimated, axis=0
            )
            correlation_values = np.sum(
                proportion_truth_centered * proportion_estimated_centered, axis=0
            ) / np.sqrt(
                np.sum(proportion_truth_centered**2, axis=0)
                * np.sum(proportion_estimated_centered**2, axis=0)
            )
            return correlation_values
        elif metric_type == "Spot":
            proportion_truth_centered = proportion_truth - np.mean(
                proportion_truth, axis=1, keepdims=True
            )
            proportion_estimated_centered = proportion_estimated - np.mean(
                proportion_estimated, axis=1, keepdims=True
            )
            correlation_values = np.sum(
                proportion_truth_centered * proportion_estimated_centered, axis=1
            ) / np.sqrt(
                np.sum(proportion_truth_centered**2, axis=1)
                * np.sum(proportion_estimated_centered**2, axis=1)
            )
            return correlation_values
        else:
            raise ValueError(f"Invalid metric type {metric_type}")

    @staticmethod
    def correct_fraction(
        proportion_truth: np.ndarray,
        proportion_estimated: np.ndarray,
        metric_type: str = "Spot",
    ) -> np.ndarray:
        assert proportion_truth.shape == proportion_estimated.shape
        correct_proportion = np.minimum(proportion_truth, proportion_estimated)
        if metric_type == "Cell type":
            return np.sum(correct_proportion, axis=0) / np.sum(proportion_truth, axis=0)
        elif metric_type == "Spot":
            return np.sum(correct_proportion, axis=1)
        else:
            raise ValueError(f"Invalid metric type {metric_type}")

    @staticmethod
    def JSD(
        proportion_truth: np.ndarray,
        proportion_estimated: np.ndarray,
        metric_type: str = "Spot",
    ) -> np.ndarray:
        """
        Jensen–Shannon divergence
        Args:
            proportion_truth: Ground truth of the cell proportion.
            proportion_estimated: Estimated proportion.
            metric_type: How the metric is calculated.
        """
        assert proportion_truth.shape == proportion_estimated.shape
        if metric_type == "Spot":
            return jensenshannon(proportion_truth, proportion_estimated, axis=1)
        elif metric_type == "Cell type":
            return jensenshannon(proportion_truth, proportion_estimated, axis=0)
        else:
            raise ValueError(f"Invalid metric type {metric_type}")

    def evaluate_metric(
        self, metric="Cosine similarity", metric_type="Spot", region=None
    ):
        """
        Evaluate the proportions based on the metric.

        Args:
            metric: Name of the metric.
            metric_type: How the metric is calculated. 'Spot': metric is calculated for each spot; 'Cell type',
                metric is calculated for each cell type; 'Individual': metric is calculated for each individual
                proportion estimation.
            region: The region that is being evaluated.
        """
        assert metric in self.metric_type_dict[metric_type]
        metric_values = []
        func = self.function_map.get(metric)
        if region is None:
            select = np.array([True] * len(self.proportion_truth))
        elif isinstance(region, list):
            select = np.array([i in region for i in self.cluster])
        else:
            select = np.array([i == region for i in self.cluster])

        if not any(select):
            raise ValueError(f"Region {region} do/does not exist.")
        elif metric_type == "Cell type" and np.sum(select) == 1:
            raise ValueError(f"Region {region} only contain(s) one spot.")
        for i in range(self.n_method):
            metric_values.append(
                func(
                    self.proportion_truth[select],
                    self.proportion_estimated_list[i][select],
                    metric_type,
                )
            )
        self.metric_dict[metric + " " + metric_type] = metric_values
        return metric_values

    def plot_metric(
        self,
        save=False,
        region=None,
        metric="Cosine similarity",
        metric_type="Spot",
        cell_types=None,
        suffix="",
        show=True,
    ):
        """Plot the box plot of each method based on the metric.

        Box number equals to the number of methods.

        Args:
            save: If true, save the figure.
            region: Regions of the tissue.
            metric: Name of the metric.
            metric_type: How the metric is calculated. 'Spot': metric is calculated for each spot; 'Cell type',
                metric is calculated for each cell type; 'Individual': metric is calculated for each individual
                proportion estimation.
            cell_types: If metric_type is 'Cell type' and cell_types is not None, then only plot the results
                corresponding to the cell_types.
            suffix: suffix of the save file.
            show: Whether to show the figure
        """
        assert metric_type == "Spot" or metric_type == "Cell type"
        assert metric in self.metric_type_dict[metric_type]
        if metric + " " + metric_type not in self.metric_dict.keys():
            self.evaluate_metric(metric=metric, metric_type=metric_type, region=region)
        region_name = ""
        if region is not None:
            region_name = (
                "_" + "+".join(region) if isinstance(region, list) else "_" + region
            )

        if metric_type == "Cell type" and cell_types is not None:
            idx = np.array(
                [np.where(self.type_list == type_)[0][0] for type_ in cell_types]
            )
            metric_values = self.metric_dict[metric + " " + metric_type]
            metric_values = [m[idx] for m in metric_values]
            metric_values = np.concatenate(metric_values)
            methods_name = np.repeat(self.methods, len(idx))
        else:
            metric_values = np.concatenate(self.metric_dict[metric + " " + metric_type])
            methods_name = np.repeat(
                self.methods, len(self.metric_dict[metric + " " + metric_type][0])
            )

        df = pd.DataFrame({metric: metric_values, "Method": methods_name})
        if show:
            plt.figure(dpi=300)
            palette = self.colors if len(self.methods) <= len(self.colors) else "Dark2"
            ax = sns.boxplot(
                data=df, y=metric, x="Method", showfliers=False, palette=palette
            )
            if metric_type == "Spot":
                ax = sns.stripplot(
                    data=df,
                    y=metric,
                    x="Method",
                    ax=ax,
                    jitter=0.2,
                    palette=palette,
                    size=1,
                )
            else:
                ax = sns.stripplot(
                    data=df,
                    y=metric,
                    x="Method",
                    ax=ax,
                    jitter=True,
                    color="black",
                    size=1.5,
                )
            for patch in ax.patches:
                r, g, b, a = patch.get_facecolor()
                patch.set_facecolor((r, g, b, 0.8))
            ax.set(xlabel="")
            sns.despine(top=True, right=True)
            plt.gca().yaxis.grid(False)
            plt.gca().xaxis.grid(False)
            plt.gca().spines["left"].set_color("black")
            plt.gca().spines["bottom"].set_color("black")
            plt.gca().tick_params(left=True, axis="y", colors="black")
            plt.gca().yaxis.set_major_locator(MaxNLocator(nbins=6))
            if save:
                plt.savefig(
                    f"{self.out_dir}figures/{metric}_{metric_type}{region_name}{suffix}.jpg",
                    dpi=500,
                    bbox_inches="tight",
                )
            plt.show()
        return df

    def plot_metric_spot_type(self, save=False, metric="Absolute error"):
        """
        Similar to plot_metric_spot, but the figures are separated for each cell type.
        """
        """
        Plot the box plot of each method based on the metric. Each value in box plot represents a spot.
        Box number equals to the number of methods.
        """
        assert metric in self.general_metric_names
        plt.figure(dpi=300)
        if metric not in self.metric_dict.keys():
            self.evaluate_metric(metric=metric)
        sns.set_palette("Dark2")
        n_spot = len(self.proportion_truth)
        metric_values = np.vstack(self.metric_dict[metric])
        metric_values = metric_values.flatten("F")
        methods_name = np.repeat(self.methods, n_spot)
        methods_name = np.tile(methods_name, self.n_type)
        cell_type = np.repeat(self.type_list, n_spot * self.n_method)

        # print(np.shape(metric_values), np.shape(methods_name), np.shape(cell_type))
        df = pd.DataFrame(
            {"metric": metric_values, "Method": methods_name, "Cell type": cell_type}
        )

        for cell_type in self.type_list:
            ax = sns.boxplot(
                data=df[df["Cell type"] == cell_type],
                y="metric",
                x="Method",
                showfliers=False,
            )
            for patch in ax.patches:
                r, g, b, a = patch.get_facecolor()
                patch.set_facecolor((r, g, b, 0.7))
            ax = sns.stripplot(
                data=df[df["Cell type"] == cell_type],
                y="metric",
                x="Method",
                ax=ax,
                jitter=0.2,
                palette="Dark2",
                size=2,
            )
            ax.set(xlabel="")
            if save:
                cell_type = "".join(x for x in cell_type if x.isalnum())
                plt.savefig(
                    f"{self.out_dir}figures/{metric} {cell_type}.jpg",
                    dpi=500,
                    bbox_inches="tight",
                )
            plt.show()
            del ax

    def plot_metric_all(self, save=False, metric="Absolute error", region=None):
        assert metric in self.general_metric_names
        plt.figure(figsize=(self.n_method * self.n_type / 4, 5), dpi=300)
        if metric not in self.metric_dict.keys():
            self.evaluate_metric(metric=metric)
        sns.set_palette("Dark2")
        region_name = ""
        if region is None:
            metric_values = np.vstack(self.metric_dict[metric])
            metric_values = metric_values.flatten("F")
            n_spot = len(self.proportion_truth)
        else:
            if isinstance(region, list):
                select = np.array([i in region for i in self.cluster])
                region_name = "_" + "+".join(region)
            else:
                select = np.array([i == region for i in self.cluster])
                region_name = "_" + region
            if not any(select):
                raise ValueError("Region must exist.")
            metric_values = [x[select] for x in self.metric_dict[metric]]
            metric_values = np.vstack(metric_values)
            metric_values = metric_values.flatten("F")
            n_spot = np.sum(select)

        methods_name = np.repeat(self.methods, n_spot)
        methods_name = np.tile(methods_name, self.n_type)
        cell_type = np.repeat(self.type_list, n_spot * self.n_method)

        df = pd.DataFrame(
            {metric: metric_values, "Method": methods_name, "Cell type": cell_type}
        )
        ax = sns.boxplot(
            data=df,
            y=metric,
            hue="Method",
            x="Cell type",
            flierprops={"marker": "o"},
            dodge=True,
            linewidth=0.6,
            fliersize=0.5,
        )
        # sns.violinplot(data=df, y=metric, hue='Method', x='Cell type')
        # ax = sns.catplot(data=df, y=metric, hue='Method', x='Cell type', kind='boxen')
        # for patch in ax.patches:
        #     r, g, b, a = patch.get_facecolor()
        #     patch.set_facecolor((r, g, b, .7))
        # ax = sns.stripplot(data=df, y=metric, x='Method', hue='Cell type', ax=ax, jitter=0.2, palette='Dark2', size=2)
        # ax.set(xlabel='')
        if save:
            plt.savefig(
                f"{self.out_dir}figures/{metric}{region_name}.tiff",
                dpi=800,
                bbox_inches="tight",
            )
        plt.show()


def decomposition(
    adata_st: anndata.AnnData,
    adata_sc: anndata.AnnData,
    key_type: str,
    cell_proportion: np.ndarray,
    save: bool = True,
    out_dir: str = "",
    threshold: float = 0.1,
    n_cell: Optional[np.ndarray] = None,
    spot_location: Optional[np.ndarray] = None,
    filtering_gene: bool = False,
    filename: str = "ST_decomposition.h5ad",
    verbose: int = 0,
    use_original_proportion: bool = False,
) -> anndata.AnnData:
    """Decompose ST.

    Args:
        adata_st: Original spatial transcriptomics data.
        adata_sc: Original single-cell data.
        key_type: The key that is used to extract cell type information from adata_sc.obs.
        cell_proportion: Proportion of each cell type obtained by the deconvolution.
        save: If True, save the generated adata_st as a file.
        out_dir: Output directory.
        threshold: If n_cell is none, discard cell types with proportion less than threshold.
        n_cell: Number of cells in each spot.
        spot_location: Coordinates of the spots.
        filtering_gene: Whether filter the genes in sc_reference.initialization.
        filename: Name of the saved file.
        verbose: Whether print the time spend.
        use_original_proportion: If the original proportion is used to estimate the iscRNA. Note that even when the
            original proportion is used, we still filter some cells in iscRNA.
    Returns:
        adata_st_decomposed: Anndata similar to scRNA, but obtained by decomposing ST.
    """
    time_start = time.time()
    type_list = sorted(list(adata_sc.obs[key_type].unique()))  # list of the cell type.
    assert len(type_list) == cell_proportion.shape[1]
    assert len(cell_proportion) == len(adata_st)
    if spot_location is not None:
        assert len(spot_location) == len(adata_st)
    n_type = len(type_list)

    cell_proportion_temp = cell_proportion.copy()
    if n_cell is None:
        select = cell_proportion_temp >= threshold
        cell_proportion_temp[cell_proportion_temp < threshold] = 0
        cell_proportion_temp = cell_proportion_temp / (
            np.sum(cell_proportion_temp, axis=1, keepdims=True) + 1e-6
        )
    else:
        cell_count = np.zeros((len(spot_location), n_type))
        for i in range(len(spot_location)):
            cell_count[i] = proportion_to_count(cell_proportion_temp[i], n_cell[i])
        cell_count = cell_count.astype(np.int32)
        select = []
        for i in range(n_type):
            select_temp = []
            for j in range(len(cell_count)):
                if cell_count[j, i] > 0:
                    select_temp += [j] * cell_count[j, i]
            select += [np.array(select_temp)]
        cell_proportion_temp = cell_count / (
            np.sum(cell_count, axis=1, keepdims=True) + 1e-6
        )
    if verbose:
        print(
            "Prepared proportion data. Time use {:.2f}".format(time.time() - time_start)
        )
        time_start = time.time()

    adata_sc_temp, adata_st_temp = sc_reference.initialization(
        adata_sc, adata_st, filtering=filtering_gene
    )
    if verbose:
        print(
            "Initialized scRNA and ST data. Time use {:.2f}".format(
                time.time() - time_start
            )
        )
        time_start = time.time()
    sc_ref = sc_reference.construct_sc_ref(adata_sc_temp, key_type=key_type)
    X = (
        adata_st_temp.X
        if type(adata_st_temp.X) is np.ndarray
        else adata_st_temp.X.toarray()
    )
    # X = np.array(adata_st_temp.X)
    if verbose:
        print(
            "Processed scRNA and ST data. Time use {:.2f}".format(
                time.time() - time_start
            )
        )
        time_start = time.time()
    if use_original_proportion:
        Y = cell_proportion[:, np.newaxis, :] * sc_ref.T
    else:
        Y = cell_proportion_temp[:, np.newaxis, :] * sc_ref.T
    Y = Y / (np.sum(Y, axis=2, keepdims=True) + 1e-10)
    Y = Y * X[:, :, np.newaxis]  # n_spot*n_gene*n_type
    if verbose:
        print(
            "Decomposition complete. Time use {:.2f}".format(time.time() - time_start)
        )
        time_start = time.time()

    cell_type = []
    ST_decompose = []
    spot_name = []
    n_cell_duplicate = []
    for i in range(n_type):
        if n_cell is not None:
            if len(select[i]) > 0:
                n_cell_duplicate += list(cell_count[select[i], i])
                ST_decompose += [Y[select[i], :, i]]
                cell_type += [type_list[i]] * np.sum(cell_count[:, i])
                spot_name += list(np.array(adata_st.obs_names)[select[i]])
        else:
            ST_decompose += [Y[select[:, i], :, i]]
            cell_type += [type_list[i]] * np.sum(select[:, i])
            spot_name += list(np.array(adata_st.obs_names)[select[:, i]])
    ST_decompose = np.vstack(ST_decompose)
    ST_decompose = (
        ST_decompose / (np.sum(ST_decompose, axis=1, keepdims=True) + 1e-10) * 1e6
    )
    adata_st_decomposed = anndata.AnnData(ST_decompose)
    adata_st_decomposed.uns = adata_st.uns
    adata_st_decomposed.var_names = adata_st_temp.var_names
    adata_st_decomposed.obs["cell_type"] = cell_type
    adata_st_decomposed.obs["spot_name"] = spot_name
    if n_cell is not None:
        adata_st_decomposed.obs["n_cell"] = n_cell_duplicate
        adata_st_decomposed.uns["cell_count"] = cell_count
    if spot_location is not None:
        if n_cell is not None:
            select = np.concatenate(select).astype(np.int32)
            location = spot_location[select]
        else:
            location = [spot_location[select[:, i], :] for i in range(n_type)]
            location = np.vstack(location)
        adata_st_decomposed.obs["location_x"] = location[:, 0]
        adata_st_decomposed.obs["location_y"] = location[:, 1]
    if verbose:
        print(
            "Constructed ST decomposition data file. Time use {:.2f}".format(
                time.time() - time_start
            )
        )
        time_start = time.time()

    if save:
        if out_dir and not os.path.exists(out_dir):
            os.mkdir(out_dir)
        out_dir = out_dir + "/" if out_dir else ""
        adata_st_decomposed.write_h5ad(out_dir + filename, compression="gzip")
        if verbose:
            print(
                "Saved file to output folder. Time use {:.2f}".format(
                    time.time() - time_start
                )
            )

    return adata_st_decomposed


def assign_type_spot(
    nucleus_df: pd.DataFrame,
    n_cell_df: pd.DataFrame,
    cell_number: np.ndarray,
    type_list: List[str],
) -> pd.DataFrame:
    """Assign the cell type to the cells inside the spot.

    Args:
        nucleus_df: Dataframe of the nucleus. Part of spotiphy.segmentation.Segmentation.
        n_cell_df: Dataframe of the number of cells in each spot. Part of spotiphy.segmentation.Segmentation.
        cell_number: Number of each cell type in each spot.
        type_list: List of the cell types.

    Returns:
        nucleus_df with assigned spot
    """
    assert len(type_list) == cell_number.shape[1]
    assert len(n_cell_df) == len(cell_number)
    cell_number = cell_number.astype(np.int32)
    if "cell_type" not in nucleus_df.columns:
        nucleus_df["cell_type"] = "unknown"
    for i in range(len(n_cell_df)):
        if n_cell_df.loc[i, "cell_count"] > 0:
            index = n_cell_df.loc[i, "Nucleus indices"]
            np.random.shuffle(index)
            j1 = 0
            for j2, n in enumerate(cell_number[i]):
                if n > 0:
                    nucleus_df.loc[index[j1 : j1 + n], "cell_type"] = type_list[j2]
                    j1 += n
    return nucleus_df


def assign_type_out(
    nucleus_df: pd.DataFrame,
    cell_proportion: np.ndarray,
    spot_centers: np.ndarray,
    type_list: List[str],
    max_distance: int = 100,
    band_width: int = 100,
) -> Tuple[pd.DataFrame, np.ndarray]:
    """Assign the cell type to the cells outside the spot.

    Args:
        nucleus_df: Dataframe of the nucleus. Part of spotiphy.segmentation.Segmentation.
        spot_centers: Centers of the spots.
        cell_proportion: Proportion of each cell type in each spot.
        type_list: List of the cell types.
        max_distance: If the distance between a nucleus and the closest spot is larger than max_distance, the cell type
            will not be assigned to this nucleus.
        band_width: Band width of the kernel.

    Returns:
        nucleus_df with assigned spot
    """
    assert len(type_list) == cell_proportion.shape[1]
    assert len(cell_proportion) == len(spot_centers)
    if "cell_type" not in nucleus_df.columns:
        nucleus_df["cell_type"] = "unknown"
    nucleus_centers = nucleus_df[["x", "y"]].values
    d = np.sum((nucleus_centers[:, :, np.newaxis] - spot_centers.T) ** 2, axis=1)
    d_min = np.min(d, axis=1)
    weights = np.exp(-d / band_width**2)
    smooth = np.sum(weights[:, :, np.newaxis] * cell_proportion, axis=1) / np.sum(
        weights, axis=1, keepdims=True
    )

    r = np.random.random(len(nucleus_centers))
    for i in range(len(nucleus_centers)):
        if d_min[i] > max_distance**2 or nucleus_df.loc[i, "in_spot"]:
            continue
        j = 0
        t_temp = r[i]
        while t_temp - smooth[i, j] > 0:
            t_temp -= smooth[i, j]
            j += 1
        nucleus_df.loc[i, "cell_type"] = type_list[j]
    return nucleus_df, smooth


def archive_assign_type_out_gp(
    nucleus_df: pd.DataFrame,
    cell_proportion: np.ndarray,
    spot_centers: np.ndarray,
    type_list: List[str],
    max_distance: int = 100,
    return_gp: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, List[GaussianProcessRegressor]]]:
    """Assign the cell type to the cells outside the spot.

    Args:
        nucleus_df: Dataframe of the nucleus. Part of spotiphy.segmentation.Segmentation.
        spot_centers: Centers of the spots.
        cell_proportion: Proportion of each cell type in each spot.
        type_list: List of the cell types.
        max_distance: If the distance between a nucleus and the closest spot is larger than max_distance, the cell type
            will not be assigned to this nucleus.
        return_gp: If return the fitted GP models.

    Returns:
        nucleus_df with assigned spot
    """
    assert len(type_list) == cell_proportion.shape[1]
    assert len(cell_proportion) == len(spot_centers)
    if "cell_type" not in nucleus_df.columns:
        nucleus_df["cell_type"] = "unknown"
    nucleus_centers = nucleus_df[["x", "y"]].values
    d = np.sum((nucleus_centers[:, :, np.newaxis] - spot_centers.T) ** 2, axis=1)
    d = np.min(d, axis=1)

    gp_list = []
    interpolation = np.zeros((len(nucleus_centers), len(type_list)))
    mean, std = np.mean(cell_proportion, axis=0), np.std(cell_proportion, axis=0)
    cell_proportion_trans = (cell_proportion - mean) / std
    print("Fitting the GP models.")
    for i in tqdm(range(len(type_list))):
        kernel = C(1.0, (1e-3, 1e3)) * RBF(100.0, (1e-1, 1e3))
        gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=1)
        gp.fit(spot_centers, cell_proportion_trans[:, i])
        gp_list += [gp]
        interpolation[:, i] = (
            gp_list[i].predict(nucleus_centers, return_std=False) * std[i] + mean[i]
        )
    interpolation[interpolation < 0] = 0
    interpolation += 1e-8
    interpolation = interpolation / np.sum(interpolation, axis=1, keepdims=True)

    print("Assigning the cell types.")
    r = np.random.random(len(nucleus_centers))
    for i in tqdm(range(len(nucleus_centers))):
        if d[i] > max_distance**2 or nucleus_df.loc[i, "in_spot"]:
            continue
        j = 0
        t_temp = r[i]
        while t_temp - interpolation[i, j] > 0:
            t_temp -= interpolation[i, j]
            j += 1
        nucleus_df.loc[i, "cell_type"] = type_list[j]
    if return_gp:
        return nucleus_df, gp_list
    else:
        return nucleus_df



import numpy as np
import matplotlib.pyplot as plt
import os
from tqdm import tqdm
import cv2 as cv
import matplotlib.patches as patches
import pandas as pd
from typing import List, Dict, Tuple, Optional, Union


class Plot_Visium:
    def __init__(
        self,
        segmentation,
        boundary_dict: Dict,
        type_list: List[str],
        colors: Optional[np.ndarray] = None,
    ):
        """
        Args:
            segmentation: Object of spotiphy.segmentation.Segmentation.
            boundary_dict: Output of function spotiphy.segmentation.cell_boundary.
            type_list: List of the cell types.
            colors: Colors of each cell type.
        """
        self.img = segmentation.img  # Original image.
        if np.max(self.img) <= 1:
            self.img = (self.img * 255).astype(np.int32)
        self.img_seg = segmentation.label  # Segmented image.
        self.n_cell_df = (
            segmentation.n_cell_df
        )  # Dataframe shown the cells in each spot.
        self.nucleus_boundary = segmentation.nucleus_boundary.astype(
            np.int32
        )  # Segmented nucleus boundaries.
        self.nucleus_df = segmentation.nucleus_df
        self.spot_radius = segmentation.spot_radius
        self.spot_centers = segmentation.spot_center

        self.nucleus_centers = self.nucleus_df[["x", "y"]].values.astype(
            np.int32
        )  # Segmented nucleus centers.
        self.img_cell = boundary_dict["img_cell"].astype(
            np.int32
        )  # Inferred cell location.
        self.type_mask = None
        self.img_boundary = boundary_dict["cell_boundary"]
        self.individual_boundary = boundary_dict["individual_boundary"]
        self.img_size = self.img_seg.shape[:2]
        self.type_list = type_list

        if colors is None:
            self.colors = np.array(
                list(plt.get_cmap("tab20b").colors)
                + list(plt.get_cmap("tab20c").colors)
            )
            self.colors = (self.colors * 255).astype(np.int32)
        else:
            self.colors = colors
        self.color_dict = {
            type_: list(self.colors[i]) for i, type_ in enumerate(type_list)
        }

    def plot(
        self,
        background: bool = False,
        cell: str = "both",
        shape: str = "cell",
        circle_size: int = 10,
        boundary: Optional[str] = None,
        save: Optional[str] = "Visium_plot.png",
        background_alpha: float = 0.5,
        spot: bool = True,
        spot_width: int = 2,
        spot_color: Tuple[int, int, int] = (0, 0, 255),
        cell_boundary_color: Tuple[int, int, int] = (100, 100, 100),
        dpi: int = 300,
    ) -> None:
        """
        Args:
            background: If show the background.
            cell: Which group of cell shapes to plot? [both, in, out]
            shape: Which shape to plot? [cell, nucleus, circle]
            circle_size: Size of the nuclei.
            boundary: Which group of cell boundary to plot? [both, in, out]
            save: If not none, save the figure to the path.
            background_alpha: Opacity of the background figure.
            spot: If plot the spot.
            spot_width: Width of the spot.
            spot_color: Color of the spot.
            cell_boundary_color: Color of the cell boundaries.
            dpi: DPI of the image plotted.
        """
        img = (
            self.img.copy() * background_alpha
            if background
            else np.zeros(self.img.shape)
        )
        img = img.astype(np.int32)

        group_dict = {
            "both": [True, False],
            "in": [True],
            "out": [False],
            None: [],
            "None": [],
        }
        type_annotation = list(self.nucleus_df["cell_type"])
        in_spot = list(self.nucleus_df["in_spot"])

        print("Adding cells.")
        if shape == "cell":
            d = {t: i for i, t in enumerate(self.type_list)}
            type_mask = np.zeros(
                (img.shape[0], img.shape[1], len(self.type_list)), dtype=bool
            )
            for i in tqdm(range(len(self.img_cell))):
                for j in range(len(self.img_cell[0])):
                    k = self.img_cell[i, j]
                    if (
                        0 < k <= len(type_annotation)
                        and type_annotation[k - 1] in d.keys()
                        and in_spot[k - 1] in group_dict[cell]
                    ):
                        type_mask[i, j, d[type_annotation[k - 1]]] = True
            for i, type_ in tqdm(enumerate(self.type_list)):
                color = np.array(self.color_dict[type_]).astype(np.int32)
                img[type_mask[:, :, i]] = color
            # img.astype(np.int32)

        for i, type_ in tqdm(enumerate(type_annotation)):
            if in_spot[i] in group_dict[cell] and type_ in self.color_dict.keys():
                color = [int(i) for i in self.color_dict[type_]]
                if shape == "circle":
                    img = cv.circle(
                        img, self.nucleus_centers[i], circle_size, color, -1
                    )
                elif shape == "nucleus":
                    boundary_temp = self.nucleus_boundary[i].reshape((-1, 1, 2))
                    img = cv.fillPoly(img, [boundary_temp], color=color)

        boundary_mask = np.zeros((img.shape[0], img.shape[1]), dtype=bool)
        for i in tqdm(range(len(type_annotation))):
            if (
                in_spot[i] in group_dict[boundary]
                and len(self.individual_boundary[i + 1]) > 0
                and type_annotation[i]
            ) in self.color_dict.keys():
                index_temp = np.array(self.individual_boundary[i + 1])
                boundary_mask[index_temp[:, 0], index_temp[:, 1]] = True
        img[boundary_mask] = cell_boundary_color

        if spot:
            print("Adding spots.")
            for i in range(len(self.spot_centers)):
                img = cv.circle(
                    img,
                    self.spot_centers[i],
                    int(self.spot_radius),
                    spot_color,
                    spot_width,
                )
        if save is not None:
            print("Saving the image.")
            img1 = img[:, :, [2, 1, 0]]
            cv.imwrite(save, img1)
        plt.figure(dpi=dpi)
        plt.imshow(img)

    def plot_legend(
        self,
        save: Optional[str] = None,
        dpi: int = 300,
        background: Tuple[int, int, int] = (0, 0, 0),
        ncol: int = 1,
        word_color: Tuple[int, int, int] = (255, 255, 255),
        fontsize: int = 6,
        horizontal_distance: int = 380,
        show: bool = False,
    ) -> None:
        nrow = len(self.type_list) // ncol
        if nrow * ncol < len(self.type_list):
            nrow += 1
        img = np.ones((nrow * 60 + 40, ncol * horizontal_distance - 30, 3)) * np.array(
            background
        )
        img = img.astype(np.int32)

        fig, ax = plt.subplots(dpi=dpi)
        plt.imshow(img)
        i = 0  # col index
        j = 0  # row index
        word_color = (word_color[0] / 255, word_color[1] / 255, word_color[2] / 255)
        for k, v in self.color_dict.items():
            ax.add_patch(
                patches.Rectangle(
                    (i * horizontal_distance + 20, j * 60 + 20),
                    100,
                    45,
                    facecolor=np.array(v) / 255,
                    edgecolor="none",
                )
            )
            ax.text(
                i * horizontal_distance + 145,
                j * 60 + 43,
                k,
                va="center",
                ha="left",
                fontsize=fontsize,
                color=word_color,
            )
            j += 1
            if j == nrow:
                j = 0
                i += 1
        ax.axis("off")
        if save is not None:
            plt.savefig(save, bbox_inches="tight", pad_inches=0)
        if show:
            plt.show()


class Plot_Xenium:
    def __init__(
        self,
        Xenium_img: np.ndarray,
        cell_boundaries: pd.DataFrame,
        nucleus_boundaries: pd.DataFrame,
        type_list: List[str],
        cell_type: List[str],
        nucleus_centers: np.ndarray,
    ):
        """
        Args:
            Xenium_img: Image of Xenium
            cell_boundaries: Coordinates of the cell boundaries on the image.
            nucleus_boundaries: Coordinates of the nucleus boundaries on the image.
            type_list: List of the cell types.
            cell_type: Annotation.
            nucleus_centers: Centers of the nuclei.
        """
        self.img = Xenium_img
        self.type_list = type_list
        self.cell_type = cell_type
        self.n_cell = len(cell_type)
        self.nucleus_centers = nucleus_centers.astype(np.int32)

        cell_boundaries_temp = cell_boundaries.iloc[:, 1:3].values.astype(np.int32)
        idx = [[] for _ in range(self.n_cell)]
        for i in tqdm(range(len(cell_boundaries))):
            cell_idx = int(cell_boundaries.iloc[i, 0] - 1)
            idx[cell_idx] += [i]
        self.cell_boundaries = [
            (
                cell_boundaries_temp[np.array(idx[i])].reshape((-1, 1, 2))
                if len(idx[i]) > 0
                else []
            )
            for i in tqdm(range(self.n_cell))
        ]

        nucleus_boundaries_temp = nucleus_boundaries.iloc[:, 1:3].values.astype(
            np.int32
        )
        idx = [[] for _ in range(self.n_cell)]
        for i in tqdm(range(len(nucleus_boundaries))):
            cell_idx = int(nucleus_boundaries.iloc[i, 0] - 1)
            idx[cell_idx] += [i]
        self.nucleus_boundaries = [
            (
                nucleus_boundaries_temp[np.array(idx[i])].reshape((-1, 1, 2))
                if len(idx[i]) > 0
                else []
            )
            for i in tqdm(range(self.n_cell))
        ]

        self.colors = np.array(
            list(plt.get_cmap("tab20b").colors) + list(plt.get_cmap("tab20c").colors)
        )
        self.colors = (self.colors * 255).astype(np.int32)
        self.color_dict = {
            type_: list(self.colors[i]) for i, type_ in enumerate(type_list)
        }

    def plot(
        self,
        background: bool = False,
        shape: str = "cell",
        save: Optional[str] = "Xenium_plot.png",
        cell_boundaries: bool = False,
        background_alpha: float = 0.8,
        circle_size: int = 10,
        cell_boundary_color: Tuple[int, int, int] = (100, 100, 100),
        cell_boundary_thickness: int = 2,
    ) -> None:
        """
        Args:
            background: If show the background.
            shape: Which shape to plot? [cell, nucleus, circle]
            save: If not none, save the figure to the path.
            cell_boundaries: If plot the cell boundaries.
            background_alpha: Opacity of the background figure.
            circle_size: Size of the circle.
            cell_boundary_color: Color of the cell boundaries.
            cell_boundary_thickness: Thickness of the cell boundaries.
        """
        img = (
            self.img.copy() * background_alpha
            if background
            else np.zeros(self.img.shape)
        )
        img = img.astype(np.int32)
        print("Adding cellls.")
        for i, type_ in tqdm(enumerate(self.cell_type)):
            if type_ in self.color_dict.keys():
                color = [int(i) for i in self.color_dict[type_]]
                if shape == "cell":
                    if len(self.cell_boundaries[i]) > 0:
                        img = cv.fillPoly(img, [self.cell_boundaries[i]], color=color)
                elif shape == "nucleus":
                    if len(self.nucleus_boundaries[i]) > 0:
                        img = cv.fillPoly(
                            img, [self.nucleus_boundaries[i]], color=color
                        )
                elif shape == "circle":
                    img = cv.circle(
                        img, self.nucleus_centers[i], circle_size, color, -1
                    )
        if cell_boundaries:
            for i, type_ in tqdm(enumerate(self.cell_type)):
                if len(self.cell_boundaries[i]) > 0:
                    img = cv.polylines(
                        img,
                        [self.cell_boundaries[i]],
                        color=cell_boundary_color,
                        thickness=cell_boundary_thickness,
                        isClosed=True,
                    )

        if save is not None:
            print("Saving the image.")
            img1 = img[:, :, [2, 1, 0]]
            cv.imwrite(save, img1)
        plt.imshow(img)



from setuptools import setup

required_packages = [
    "squidpy==1.2.3",
    "scanpy==1.9.3",
    "scikit-learn==1.2.2",
    "scipy==1.9.1",
    "pandas==2.0.0",
    "numpy==1.22.4",
    "matplotlib",
    "seaborn",
    "tqdm",
    # "charset-normalizer==3.1.0",
    "charset-normalizer",
    # "tensorflow==2.12.0",
    "stardist==0.8.3",
    # "torch",
    "pyro-ppl==1.8.4",
    "opencv-python>=4.7",
    "jupyter",
]

setup(
    name="spotiphy",
    version="0.2.10",
    packages=["spotiphy"],
    url="https://github.com/jyyulab/Spotiphy",
    license="Apache-2.0",
    author="Ziqian Zheng, Jiyuan Yang",
    author_email="zzheng92@wisc.edu",
    description="An integrated pipeline designed to deconvolute and decompose spatial transcriptomics data, "
    "and produce pseudo single-cell resolution images.",
    python_requires=">=3.9",
    install_requires=required_packages,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
)



import anndata
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from tqdm import tqdm
from scipy import stats
import time
from typing import List, Dict, Tuple, Union
from typing import List, Dict, Union


def initialization(
    adata_sc: anndata.AnnData,
    adata_st: anndata.AnnData,
    min_genes: int = 200,
    min_cells: int = 200,
    min_std: float = 20,
    normalize_st: np.ndarray = None,
    filtering: bool = True,
    verbose: int = 0,
) -> Tuple[anndata.AnnData, anndata.AnnData]:
    """
    Filter single cell and spatial data, normalize to CPM, and optionally filter genes/cells.

    Args:
        adata_sc: Single cell data (AnnData).
        adata_st: Spatial data (AnnData).
        min_genes: Minimum genes per cell for filtering.
        min_cells: Minimum cells per gene for filtering.
        min_std: Minimum std for gene filtering after CPM normalization.
        normalize_st: If None, spatial data is normalized to one million. If ndarray, represents cell counts per spot.
        filtering: Whether to filter genes/cells in adata_sc.
        verbose: Verbosity level.

    Returns:
        Tuple of filtered and normalized (adata_sc, adata_st).
    """
    time_start = time.time()
    X = adata_sc.X if isinstance(adata_sc.X, np.ndarray) else adata_sc.X.toarray()
    Y = adata_st.X if isinstance(adata_st.X, np.ndarray) else adata_st.X.toarray()
    if verbose == 1:
        print(
            f"Convert expression matrix to array: {np.round(time.time()-time_start, 2)}s"
        )
        time_start = time.time()

    # CPM normalization
    X = X * 1e6 / X.sum(axis=1, keepdims=True)
    if normalize_st is None:
        Y = Y * 1e6 / Y.sum(axis=1, keepdims=True)
    else:
        if not isinstance(normalize_st, np.ndarray) or normalize_st.shape != (
            len(adata_st),
        ):
            raise ValueError("normalize_st must be a numpy array with shape (n_spots,)")
        Y = Y * 1e6 / Y.sum(axis=1, keepdims=True) * normalize_st[:, np.newaxis]
    if verbose == 1:
        print(f"Normalization: {np.round(time.time()-time_start, 2)}s")
        time_start = time.time()
    adata_st.X = Y

    # Initial filtering
    if filtering:
        cell_f1 = np.sum(X > 0, axis=1) > min_genes
        gene_f1 = np.sum(X > 0, axis=0) > min_cells
        gene_f2 = np.std(X, axis=0) > min_std
        adata_sc.X = X
        adata_sc = adata_sc[cell_f1]
        adata_sc = adata_sc[:, np.logical_and(gene_f1, gene_f2)]
        if verbose == 1:
            print(f"Filtering: {np.round(time.time()-time_start, 2)}s")
            time_start = time.time()
    else:
        adata_sc.X = X

    # Find the intersection of genes
    common_genes = list(set(adata_st.var_names).intersection(set(adata_sc.var_names)))
    adata_sc = adata_sc[:, common_genes]
    adata_st = adata_st[:, common_genes]
    if verbose == 1:
        print(f"Find common genes: {np.round(time.time()-time_start, 2)}s")
    return adata_sc, adata_st


def marker_selection(
    adata_sc: anndata.AnnData,
    key_type: str,
    threshold_cover: float = 0.6,
    threshold_p: float = 0.1,
    threshold_fold: float = 1.5,
    n_select: int = 40,
    verbose: int = 0,
    return_dict: bool = False,
    q: float = 0.0,
) -> Union[List[str], Dict[str, List[str]]]:
    """
    Find marker genes based on pairwise ratio test.

    Args:
        adata_sc: scRNA data (AnnData).
        key_type: Key for cell type in adata_sc.obs.
        threshold_cover: Min proportion of non-zero reads in cell type.
        threshold_p: Max p-value for marker gene.
        threshold_fold: Min fold change for marker gene.
        n_select: Number of marker genes per cell type.
        verbose: Verbosity level.
        return_dict: If True, return dict of marker genes by cell type.
        q: Quantile for fold-change selection.

    Returns:
        List or dict of marker genes.
    """
    X = adata_sc.X if isinstance(adata_sc.X, np.ndarray) else adata_sc.X.toarray()

    # Derive mean and std matrix
    type_list = sorted(list(adata_sc.obs[key_type].unique()))  # list of the cell type.
    n_gene, n_type = adata_sc.shape[1], len(type_list)
    expression_mu = np.zeros(
        (n_type, n_gene)
    )  # Mean expression of each gene in each cell type.
    expression_sd = np.zeros(
        (n_type, n_gene)
    )  # Standard deviation of expression of each gene in each cell type.
    n_cell_by_type = np.zeros(n_type)
    data_type = []  # The expression data categorized by cell types.
    for i in range(n_type):
        data_type.append(X[adata_sc.obs[key_type] == type_list[i]])
        expression_mu[i] = np.mean(data_type[i], axis=0)
        expression_sd[i] = np.std(data_type[i], axis=0)
        n_cell_by_type[i] = len(data_type[i])
    del X
    expression_sd = expression_sd + 1e-10
    expression_mu = expression_mu + 1e-10

    # t test
    fold_change = np.zeros((n_type, n_gene))
    p_value = np.zeros((n_type, n_gene))
    type_index_max = np.argmax(
        expression_mu, axis=0
    )  # Cell type index with the maximum mean expression of each gene
    for i in range(n_gene):
        mu0 = expression_mu[type_index_max[i], i]
        sd0 = expression_sd[type_index_max[i], i]
        n0 = n_cell_by_type[type_index_max[i]]
        A = sd0**2 / n0 + expression_sd[:, i] ** 2 / n_cell_by_type
        B = (sd0**2 / n0) ** 2 / (n0 - 1) + (
            expression_sd[:, i] ** 2 / n_cell_by_type
        ) ** 2 / (n_cell_by_type - 1)
        t_stat = (mu0 - expression_mu[:, i]) / np.sqrt(A)
        fold_change[:, i] = mu0 / expression_mu[:, i]
        df = A**2 / B
        p_value[:, i] = stats.t.sf(abs(t_stat), df)

    # determine the marker genes
    p_value_sort = np.sort(p_value, axis=0)
    fold_change_sort = np.sort(fold_change, axis=0)
    gene_name = np.array(adata_sc.var_names)
    marker_gene = dict() if return_dict else []
    for i in range(n_type):
        # fraction of non-zero reads in current datatype
        cover_fraction = (
            np.sum(data_type[i][:, type_index_max == i] > 0, axis=0) / n_cell_by_type[i]
        )
        p_value_temp = p_value_sort[-2, type_index_max == i]  # second-largest p-value
        # fold_change_temp = fold_change_sort[1, type_index_max == i]  # second-smallest fold change
        fold_change_temp = fold_change_sort[
            max(1, int(np.round(q * (n_type - 1)))), type_index_max == i
        ]
        selected = np.logical_and(
            cover_fraction >= threshold_cover, p_value_temp < threshold_p
        )
        selected = np.logical_and(fold_change_temp >= threshold_fold, selected)
        gene_name_temp = gene_name[type_index_max == i][selected]

        fold_change_temp = fold_change_temp[selected]
        selected_gene_idx = np.argsort(fold_change_temp)[::-1][:n_select]
        selected_gene = gene_name_temp[selected_gene_idx]

        if return_dict:
            marker_gene[type_list[i]] = list(selected_gene)
        else:
            marker_gene.extend(list(selected_gene))
        if verbose == 1:
            print(f"{type_list[i]}: {len(list(selected_gene)):d}")
        elif len(list(selected_gene)) < 5:
            print(
                f"Warning: Only {len(list(selected_gene)):d} genes are selected for {type_list[i]}."
            )
    return marker_gene


def construct_sc_ref(adata_sc: anndata.AnnData, key_type: str) -> np.ndarray:
    """
    Construct the scRNA reference from scRNA data.

    Args:
        adata_sc: scRNA data (AnnData).
        key_type: Key for cell type in adata_sc.obs.

    Returns:
        scRNA reference (np.ndarray, shape n_type x n_gene).
    """
    type_list = sorted(list(adata_sc.obs[key_type].unique()))
    n_gene, n_type = adata_sc.shape[1], len(type_list)
    sc_ref = np.zeros((n_type, n_gene))
    X = np.array(adata_sc.X)
    for i, cell_type in tqdm(enumerate(type_list)):
        sc_X_temp = np.sum(X[adata_sc.obs[key_type] == cell_type], axis=0)
        sc_ref[i] = sc_X_temp / np.sum(sc_X_temp)
    return sc_ref


def plot_sc_ref(
    sc_ref: np.ndarray,
    type_list: List[str],
    fig_size: Tuple[int, int] = (10, 4),
    dpi: int = 300,
) -> None:
    """
    Plot the heatmap of the single cell reference.

    Args:
        sc_ref: scRNA reference (np.ndarray, n_type x n_gene).
        type_list: List of cell types.
        fig_size: Figure size.
        dpi: Dots per inch.
    """
    fig_size_adjust = (fig_size[0], fig_size[1] * sc_ref.shape[0] / 20)
    plt.figure(figsize=fig_size_adjust, dpi=dpi)
    sc_ref_df = pd.DataFrame(sc_ref, index=type_list)
    sns.heatmap(sc_ref_df, robust=True)
    plt.show()


def plot_heatmap(
    adata_sc: anndata.AnnData,
    key_type: str,
    fig_size: Tuple[int, int] = (10, 4),
    dpi: int = 300,
    save: bool = False,
    out_dir: str = "",
) -> None:
    """
    Plot the heatmap of the mean expression.

    Args:
        adata_sc: scRNA data (AnnData).
        key_type: Key for cell type in adata_sc.obs.
        fig_size: Figure size.
        dpi: Dots per inch.
        save: Whether to save the heatmap.
        out_dir: Output directory.
    """
    X = adata_sc.X if isinstance(adata_sc.X, np.ndarray) else adata_sc.X.toarray()
    type_list = sorted(list(adata_sc.obs[key_type].unique()))  # list of the cell type.
    n_gene, n_type = adata_sc.shape[1], len(type_list)
    expression_mu = np.zeros(
        (n_type, n_gene)
    )  # Mean expression of each gene in each cell type.

    for i in range(n_type):
        data_temp = X[adata_sc.obs[key_type] == type_list[i]]
        expression_mu[i] = np.mean(data_temp, axis=0)
    del X

    expression_mu = expression_mu / np.max(expression_mu, axis=0)
    fig_size_adjust = (fig_size[0], fig_size[1] * n_type / 20)
    plt.figure(figsize=fig_size_adjust, dpi=dpi)
    sc_ref_df = pd.DataFrame(expression_mu, index=type_list)
    sns.heatmap(sc_ref_df, robust=True)
    if save:
        plt.savefig(out_dir + "heatmap.jpg", bbox_inches="tight")
    plt.show()
